import org.springframework.beans.BeanUtils;

import java.lang.annotation.Annotation
import java.lang.reflect.Method
import java.lang.reflect.Modifier
import java.lang.reflect.ParameterizedType 
import java.lang.reflect.Type 
import java.util.Map;
import java.util.regex.Matcher 

import grails.plugins.dto.DTO
import grails.plugins.dto.annotation.DtoIdOnly;
import grails.plugins.dto.annotation.DtoIgnore;

import org.codehaus.groovy.grails.plugins.dto.BetterGrailsDtoGenerator
import org.dozer.spring.DozerBeanMapperFactoryBean
import org.springframework.context.ApplicationContext

class DtoGrailsPlugin {
    // the plugin version
    def version = "0.2.4.e7"
    // the version or versions of Grails the plugin is designed for
    def grailsVersion = "1.1 > *"
    // the other plugins this plugin depends on
    def dependsOn = [:]
    def loadAfter = [ "converters" ]

    // resources that are excluded from plugin packaging
    def pluginExcludes = [
            "grails-app/views/error.gsp"
    ]

    // TODO Fill in these fields
    def author = "Peter Ledbrook"
    def authorEmail = ""
    def title = "DTO plugin for Grails"
    def description = """\\
Manually creating and managing DTO objects for your domain classes is
labour intensive and error-prone. This plugin aims to simplify DTOs
by automatically generating them and providing a mechanism to easily
map domain class instances to DTO instances.
"""

    // URL to the plugin's documentation
    def documentation = "http://grails.org/plugin/dto"

    def doWithSpring = {
        // Create the DTO generator bean.
        if (application.config.grails.generate.indent) {
            dtoGenerator(BetterGrailsDtoGenerator, true, application.config.grails.generate.indent)
        }
        else {
            dtoGenerator(BetterGrailsDtoGenerator)
        }

        dozerMapper(DozerBeanMapperFactoryBean) {
            if (application.config.dto.mapping.files) {
                mappingFiles = application.config.dto.mapping.files
            }
        }
    }

    def doWithDynamicMethods = { final ctx ->
        // Add "as DTO" and toDTO() to domain classes.
        for (dc in application.domainClasses) {
            addDtoMethods(dc.metaClass, ctx, application)
        }

        // Add the toDTO(Class) method to collections.
        Collection.metaClass.toDTO = { obj ->
            // Find out what class the target collection should contain.
            def containedClass = obj instanceof Class ? obj : obj.getClass()

            // Next create a collection of the appropriate type.
            def clazz = delegate.getClass()
            if (SortedSet.isAssignableFrom(clazz)) {
                obj = new TreeSet()
            }
            else if (Set.isAssignableFrom(clazz)) {
                obj = new HashSet()
            }
            else {
                obj = new ArrayList(delegate.size())
            }

            // Finally, add the individual DTOs to the new collection.
            final mapper = ctx.getBean("dozerMapper")
            delegate.each { obj << mapper.map(it, containedClass) }
            return obj
        }
    }

    def doWithApplicationContext = { final ctx ->
    }

    def onChange = { event ->
        // TODO Implement code that is executed when any artefact that this plugin is
        // watching is modified and reloaded. The event contains: event.source,
        // event.application, event.manager, event.ctx, and event.plugin.
    }

    def onConfigChange = { event ->
        // TODO Implement code that is executed when the project configuration changes.
        // The event is the same as for 'onChange'.
    }

    private addDtoMethods(final MetaClass mc, final ApplicationContext ctx, final def application) {
        // First add the "as DTO".
        final originalAsType = mc.getMetaMethod("asType", [ Class ] as Object[])
        mc.asType = { Class clazz ->
            if (DTO == clazz) {
                // Do the DTO conversion.
                return mapDomainInstance(ctx, delegate)
            }
            else {
                // Use the original asType implementation.
                return originalAsType.invoke(delegate, [ clazz ] as Object[])
            }
        }

        // Then the toDTO() method.
        mc.toDTO = {->
            return mapDomainInstance(ctx, delegate)
        }

        mc.toDTO = { Class clazz ->
            // Convert the domain instance to a DTO.
            def mapper = ctx.getBean("dozerMapper")
            return mapper.map(delegate, clazz)
        }

        mc.static.fromDTO = { def dto ->
            // Convert the DTO instance to a domain instance
            mapDtoInstance(application, dto, delegate)
//            
//            def mapper = ctx.getBean("dozerMapper")
//            return mapper.map(delegate, clazz)
        }
    }

    /**
     * Uses the Dozer mapper to map a domain instance to its corresponding
     * DTO.
     * @param ctx The Spring application context containing the Dozer
     * mapper.
     * @param obj The domain instance to map.
     * @return The DTO corresponding to the given domain instance.
     */
    private mapDomainInstance(ctx, obj) {
        // Get the appropriate DTO class for this domain instance.
        def dtoClassName = obj.getClass().name + "DTO"
        def dtoClass = obj.getClass().classLoader.loadClass(dtoClassName)

        // Now convert the domain instance to a DTO.
        def mapper = ctx.getBean("dozerMapper")
        return mapper.map(obj, dtoClass)
    }

    
    def processProperty = { application, domainInstance, dto, prop, idOnlyFields, ignoreFields, alreadyHandledDomainClasses ->
        println "Processing Property: prop=$prop.name, domainInstance class=${domainInstance.getClass().getName()}, dto class=${dto.getClass().getName()}"
        
        if (ignoreFields.contains(prop.name)) {
            println "Ignoring ${prop.name}"
            return
        } else {
            println "$prop.name ist kein ignoreField"
        }
        
        if (idOnlyFields.contains(prop.name)) {
            println "Finding value for dto association ${prop.name}Id: "+dto."${prop.name}Id"
            def propDbValue = prop.referencedDomainClass.clazz.get(dto."${prop.name}Id")
            println "Found ${prop.referencedDomainClass.clazz.name} with ${prop.name}Id="+dto."${prop.name}Id"
            domainInstance."${prop.name}" = propDbValue
        } else {
            println "$prop.name ist kein idOnlyField"
        }
        
        def propType = prop.type

        if (prop.referencedPropertyType == propType) {
            if (prop.association) {
                if (dto."${prop.name}" != null) {
                    println "getting value of $propType for id "+(dto."${prop.name}".id)
                    def propDbValue = propType.get(dto."${prop.name}".id)
                    println "Found value of $propType for id "+(dto."${prop.name}".id)
                    domainInstance."${prop.name}" = propDbValue
                } else {
                    println "value of dto.${prop.name} is null"
                }
            } else {
                println "Normal value for $propType = "+(dto."${prop.name}")
                domainInstance."${prop.name}" = dto."${prop.name}"
            }
        }
        else {
            println "$propType is a collection type"
            def dtoColl = dto."${prop.name}"
            
            println "dto collection = "+(dtoColl!=null?dtoColl.size()+" Eintraege":"null")

            if (dtoColl instanceof Collection) {
            
                def domainColl = domainInstance."${prop.name}"
                println "domain collection = "+(domainColl!=null?domainColl.size()+" Eintraege":"null")
                if (domainColl == null) {
                    println "domain collection is null and dto collection is not, therefore creating new $propType"
                    domainColl = BeanUtils.instantiateClass(propType)
                    domainInstance."${prop.name}" = domainColl
                } else if (domainColl.size() > 0) {
                    println "clearing domain collection"
                    domainColl.clear()
                } else {
                    println "domain collection is empty"
                }
            
            
                if (prop.isBasicCollectionType()) {
                    println "copying primitive collection entries"
                    dtoColl.each { domainColl << it }
                    println "domain collection after copying: "+domainColl
                    
                } else {
                        
                    Class entryClass = null
                    if (Set.isAssignableFrom(propType) || List.isAssignableFrom(propType)) {
                        def tp = propType.getTypeParameters()
                        println "tp=$tp"
                        println "propType als String=$propType"
                        println "referencedPropType =${prop.referencedPropertyType.name}"
                    }
                    
                    for (nextDto in dtoColl) {
                        println "mapping next dto in collection: "+nextDto
                        if (nextDto != null) {
                        
                            def nextDomainInstance = nextDto.id ? prop.referencedPropertyType.get(nextDto.id) : BeanUtils.instantiateClass(prop.referencedPropertyType)
                            def dc = application.getDomainClass(prop.referencedPropertyType.name)
                            List nextIdOnlyFields = []
                            List nextIgnoreFields = []
                            findIdOnlyAndIgnoreFields(nextIdOnlyFields, nextIgnoreFields, dc)
                            
                            dc.clazz.declaredFields.each { nextField ->
                                if (nextField && nextField.getAnnotation(DtoIdOnly.class)) {
                                    nextIdOnlyFields << nextField.name
                                }
                            }

                            def bidirectionalPropName = null
                            dc.persistentProperties.each { nextProp ->
                                if (nextProp.otherSide == prop) {
                                    bidirectionalPropName = nextProp.name
                                } else if (!alreadyHandledDomainClasses.contains(nextProp.type.name)) {
                                    alreadyHandledDomainClasses << dc.clazz.name
                                    processProperty(application, nextDomainInstance, nextDto, nextProp, nextIdOnlyFields, nextIgnoreFields, alreadyHandledDomainClasses)
                                }
                            }
                            domainColl << nextDomainInstance
                            if (bidirectionalPropName) {
                                nextDomainInstance."$bidirectionalPropName" = domainInstance
                            }
                        }
                    }
                
                }
            }
        }
    }
    

    
    private mapDtoInstance(final def application, def dto, def domainClass) {
    
        if (dto) {
            def domainInstance = dto.id ? domainClass.get(dto.id) : BeanUtils.instantiateClass(domainClass)
            def dc = application.getDomainClass(domainClass.name)
            if (dc) {
                List ignoreFields = []
                List idOnlyFields = []
                findIdOnlyAndIgnoreFields(idOnlyFields, ignoreFields, dc)

                def alreadyHandledDomainClasses = []
                alreadyHandledDomainClasses << dc.clazz.name
                dc.persistentProperties.each { prop ->
                    processProperty(application, domainInstance, dto, prop, idOnlyFields, ignoreFields, alreadyHandledDomainClasses)
                }
                
                return domainInstance
                
            } else {
                println "ERROR: no DomainClass found for name ${delegate.name}"
            }
        } else {
            println "ERROR: dto is null"
        }
    }

    private void findIdOnlyAndIgnoreFields(List idOnlyFields, List ignoreFields, dc) {
        dc.clazz.declaredFields.each { field ->
            if (field) {
                if (field.getAnnotation(DtoIdOnly.class)) {
                    idOnlyFields << field.name
                } else if (field.getAnnotation(DtoIgnore.class)) {
                    ignoreFields << field.name
                }
            }
        }
        println "found idOnlyFields: $idOnlyFields"
        println "found ignoreFields: $ignoreFields"
    }
}

