package org.codehaus.groovy.grails.plugins.dto

import grails.plugins.dto.annotation.DtoEquality
import grails.plugins.dto.annotation.DtoIdOnly
import grails.plugins.dto.annotation.DtoIgnore
import grails.plugins.dto.annotation.DtoImplements
import grails.plugins.dto.annotation.DtoNoVersion
import grails.plugins.dto.annotation.DtoSummary
import grails.plugins.dto.annotation.DtoSummaryImplements

import org.codehaus.groovy.grails.commons.GrailsDomainClass


/**
 * Default implementation of the DTO generator. It inspects the domain
 * class descriptors and uses that information to write out the DTO
 * files directly. You can specify whether the generated files have
 * native line endings or simply '\n'.
 *
 * @author Peter Ledbrook
 */
class BetterGrailsDtoGenerator {
    private final String eol
    private final String indent

    private processed

    Map packageTransforms = [:]

    /**
     * Creates a generator.
     * @param useNativeEol The generator creates files with native line
     * endings if this is <code>true</code>, otherwise it uses '\n'.
     * Default value is <code>true</code>.
     * @param indent The string to use for indenting. Defaults to 4
     * spaces.
     */
    BetterGrailsDtoGenerator(boolean useNativeEol = true, String indent = "    ") {
        if (useNativeEol) eol = System.getProperty("line.separator")
        else eol = "\n"

        this.indent = indent
    }
	
    /**
     * Generates the DTO files for the given domain class.
     * @param dc The domain class to generate a DTO for.
     * @param rootDir The root directory where the DTO files will be
     * generated, e.g. "src/java". Package directories will be created
     * within this directory as needed.
     * @param recursive If <code>true</code>, the method will generate
     * DTOs for all related domain classes too. Otherwise only the one
     * DTO for 'dc' will be created.
     */
    void generate(GrailsDomainClass dc, File rootDir, boolean recursive) {
        processed = [] as Set
        generateInternal(dc, rootDir, recursive)
    }

    /**
     * Generates the DTO file for the given domain class, without also
     * generating DTOs for related domain classes. Rather than write
     * the DTO to a file, this method will write it out to the given
     * Writer.
     * @param dc The domain class to generate a DTO for.
     * @param writer The writer to use when generating the DTO class.
     */
    Set generateNoRecurse(GrailsDomainClass dc, Writer writer) {
        final dcPkg = dc.clazz.package?.name
        return generateNoRecurseInternal(dc, writer, getTargetPackage(dcPkg))
    }

    private void generateInternal(GrailsDomainClass dc, File rootDir, boolean recursive) {
        final targetPkg = getTargetPackage(dc.clazz.package?.name)
        def dtoFile = getDtoFile(rootDir, dc, targetPkg)
        def summaryFile = getSummaryFile(rootDir, dc, targetPkg)
        dtoFile.parentFile.mkdirs()
		summaryFile.parentFile.mkdirs()
		
        def writer = new BufferedWriter(new FileWriter(dtoFile))
        try {
            def relations = generateNoRecurseInternal(dc, writer, summaryFile, targetPkg)
            processed << dc.clazz

            if (recursive && relations) {
                relations.each { rel ->
                    if (!processed.contains(rel.clazz)) {
                        generateInternal(rel, rootDir, true)
                    }
                }
            }
        }
        finally {
            if (writer) writer.close()
        }
    }

    private Set generateNoRecurseInternal(GrailsDomainClass dc, Writer writer, File summaryFile, String targetPkg) {
        // Deal with the persistent properties.
        def imports = []
        def fields = []
        def relations = [] as Set
        
        List idOnlyFields = getIdOnlyFields(dc.getClazz())
        List dtoImplements = getDtoImplements(dc.getClazz())
		List ignoredFields = getIgnoredFields(dc.getClazz())
		boolean addVersionAttribute = dc.version && includeVersion(dc.getClazz())
		
        def processProperty = { prop ->
			
			if (!ignoredFields.contains(prop.name)) {
				def propType = prop.type
				
				def field = [ name: prop.name ]
				if (idOnlyFields.contains(prop.name)) {
					field["idName"] = prop.name+"Id"
					field["idTypeString"] = prop.referencedDomainClass.identifier.type.simpleName
					println "field=$field"
				} else {
					println "$prop.name ist kein idOnly field"
				}
	
				if (prop.referencedPropertyType == propType) {
					field["typeString"] = propType.simpleName + (prop.association ? "DTO" : "")
					addImportIfNecessary(imports, targetPkg, propType, prop.association)
				}
				else {
					field["typeString"] = propType.simpleName + '<' + prop.referencedPropertyType.simpleName + (prop.association ? "DTO" : "") + '>'
					addImportIfNecessary(imports, targetPkg, propType, false)
					addImportIfNecessary(imports, targetPkg, prop.referencedPropertyType, prop.association)
				}
	
				// Store the reference domain class if this property is
				// an association. This is so that we can return a set
				// of related domain classes.
				if (prop.association) relations << prop.referencedDomainClass
	
				fields << field
			}
        }

        processProperty.call(dc.identifier)
        processProperty.call(dc.version)
        dc.persistentProperties.each(processProperty)

        // Start with the package line.
        if (targetPkg) {
            writer.write "package ${targetPkg};${eol}${eol}"
        }

        // Now add any required imports.
        if (imports) {
            imports.unique().sort().each { str ->
                writer.write "import ${str};${eol}"
            }
            writer.write eol
        }

        // Next, the class declaration.
        String dtoSimpleClassName = "${dc.shortName}DTO"
        String dtoImplementsString = "grails.plugins.dto.DTO" + (dtoImplements ? (", "+dtoImplements.join(', ')) : "") 
        
        writer.write "public class ${dtoSimpleClassName} implements $dtoImplementsString {${eol}"
        
        // A serialUID, since DTOs are serialisable.
        writer.write "${indent}private static final long serialVersionUID = 1L;${eol}${eol}"

        List equalityFields = getEqualityFields(dc.getClazz())

        // The private fields.
        fields.each { field ->
            if (field.idName) { //idOnlyFields.contains(field.name)
                writer.write "${indent}private ${field.idTypeString} ${field.idName};${eol}"
            } else {
                writer.write "${indent}private ${field.typeString} ${field.name};${eol}"
            }
        }

        // The getters and setters.
        writer.write eol
        fields.each { field ->
            def propSuffix = new StringBuilder(field.name)
            propSuffix.setCharAt(0, Character.toUpperCase(propSuffix.charAt(0)))
            propSuffix = propSuffix.toString()

            if (field.idName) { //idOnlyFields.contains(field.name)
                writer.write "${indent}public ${field.typeString} get${propSuffix}() { return null; }${eol}"
                writer.write "${indent}public void set${propSuffix}(${field.typeString} ${field.name}) { this.${field.idName} = (${field.name}!=null) ? ${field.name}.getId() : null; }${eol}"
                
                def idPropSuffix = new StringBuilder(field.idName)
                idPropSuffix.setCharAt(0, Character.toUpperCase(idPropSuffix.charAt(0)))
                idPropSuffix = idPropSuffix.toString()
                writer.write "${indent}public ${field.idTypeString} get${idPropSuffix}() { return ${field.idName}; }${eol}"
                writer.write "${indent}public void set${idPropSuffix}(${field.idTypeString} ${field.idName}) { this.${field.idName} = ${field.idName}; }${eol}"
            } else if (field.typeString == 'boolean') {
                writer.write "${indent}public ${field.typeString} is${propSuffix}() { return ${field.name}; }${eol}"
                writer.write "${indent}public void set${propSuffix}(${field.typeString} ${field.name}) { this.${field.name} = ${field.name}; }${eol}"
            } else {
                writer.write "${indent}public ${field.typeString} get${propSuffix}() { return ${field.name}; }${eol}"
                writer.write "${indent}public void set${propSuffix}(${field.typeString} ${field.name}) { this.${field.name} = ${field.name}; }${eol}"
            }

        }

        if (equalityFields.size() == 1) {
            writeSimpleHashCode(equalityFields[0], writer)
            writeEquals(equalityFields, dtoSimpleClassName, writer)
            def foundField = fields.find{ field -> field.name==equalityFields[0]}
            String equalityFieldType = foundField ? foundField.typeString : "Error: could not determine type of field '${equalityFields[0]}'"
            writeSimpleToString(equalityFields[0], equalityFieldType, writer)
        } else if (equalityFields.size() > 1) {
            writeComplexHashCode(equalityFields[0], writer)
            writeEquals(equalityFields, dtoSimpleClassName, writer)
            writeComplexToString(fields, dtoSimpleClassName, writer)
        } else {
            writeComplexToString(fields, dtoSimpleClassName, writer)
        }
                 

        // Class terminator.
        writer.write "}${eol}"

        // All done. Make sure all data has been pushed to the destination
        // before we leave.
        writer.flush()

        return relations
    }
	
	private Set generateSummaryFile(GrailsDomainClass dc, Writer writer, String targetPkg) {
		// Deal with the persistent properties.
		def imports = []
		def fields = []
		def relations = [] as Set
		
		List sumaryFields = getSummaryFields(dc.getClazz())
		List dtoSummaryImplements = getDtoSummaryImplements(dc.getClazz())
		boolean addVersionAttribute = dc.version && includeVersion(dc.getClazz())
		
		def processProperty = { prop ->
			
			if (summaryFields.contains(prop.name)) {
				def propType = prop.type
				
				def field = [ name: prop.name ]
				if (summaryFields.contains(prop.name)) {
					field["idName"] = prop.name+"Id"
					field["idTypeString"] = prop.referencedDomainClass.identifier.type.simpleName
					println "field=$field"
				} else {
					println "$prop.name ist kein idOnly field"
				}
	
				if (prop.referencedPropertyType == propType) {
					field["typeString"] = propType.simpleName + (prop.association ? "DTO" : "")
					addImportIfNecessary(imports, targetPkg, propType, prop.association)
				}
				else {
					field["typeString"] = propType.simpleName + '<' + prop.referencedPropertyType.simpleName + (prop.association ? "DTO" : "") + '>'
					addImportIfNecessary(imports, targetPkg, propType, false)
					addImportIfNecessary(imports, targetPkg, prop.referencedPropertyType, prop.association)
				}
	
				// Store the reference domain class if this property is
				// an association. This is so that we can return a set
				// of related domain classes.
				if (prop.association) relations << prop.referencedDomainClass
	
				fields << field
			}
		}

		processProperty.call(dc.identifier)
		dc.persistentProperties.each(processProperty)

		// Start with the package line.
		if (targetPkg) {
			writer.write "package ${targetPkg};${eol}${eol}"
		}

		// Now add any required imports.
		if (imports) {
			imports.unique().sort().each { str ->
				writer.write "import ${str};${eol}"
			}
			writer.write eol
		}

		// Next, the class declaration.
		String dtoSimpleClassName = "${dc.shortName}DTO"
		String dtoImplementsString = "grails.plugins.dto.DTO" + (dtoImplements ? (", "+dtoImplements.join(', ')) : "")
		
		writer.write "public class ${dtoSimpleClassName} implements $dtoImplementsString {${eol}"
		
		// A serialUID, since DTOs are serialisable.
		writer.write "${indent}private static final long serialVersionUID = 1L;${eol}${eol}"

		List equalityFields = getEqualityFields(dc.getClazz())

		// The private fields.
		fields.each { field ->
			if (field.idName) { //idOnlyFields.contains(field.name)
				writer.write "${indent}private ${field.idTypeString} ${field.idName};${eol}"
			} else {
				writer.write "${indent}private ${field.typeString} ${field.name};${eol}"
			}
		}

		// The getters and setters.
		writer.write eol
		fields.each { field ->
			def propSuffix = new StringBuilder(field.name)
			propSuffix.setCharAt(0, Character.toUpperCase(propSuffix.charAt(0)))
			propSuffix = propSuffix.toString()

			if (field.idName) { //idOnlyFields.contains(field.name)
				writer.write "${indent}public ${field.typeString} get${propSuffix}() { return null; }${eol}"
				writer.write "${indent}public void set${propSuffix}(${field.typeString} ${field.name}) { this.${field.idName} = (${field.name}!=null) ? ${field.name}.getId() : null; }${eol}"
				
				def idPropSuffix = new StringBuilder(field.idName)
				idPropSuffix.setCharAt(0, Character.toUpperCase(idPropSuffix.charAt(0)))
				idPropSuffix = idPropSuffix.toString()
				writer.write "${indent}public ${field.idTypeString} get${idPropSuffix}() { return ${field.idName}; }${eol}"
				writer.write "${indent}public void set${idPropSuffix}(${field.idTypeString} ${field.idName}) { this.${field.idName} = ${field.idName}; }${eol}"
			} else {
				writer.write "${indent}public ${field.typeString} get${propSuffix}() { return ${field.name}; }${eol}"
				writer.write "${indent}public void set${propSuffix}(${field.typeString} ${field.name}) { this.${field.name} = ${field.name}; }${eol}"
			}

		}

		if (equalityFields.size() == 1) {
			writeSimpleHashCode(equalityFields[0], writer)
			writeEquals(equalityFields, dtoSimpleClassName, writer)
			def foundField = fields.find{ field -> field.name==equalityFields[0]}
			String equalityFieldType = foundField ? foundField.typeString : "Error: could not determine type of field '${equalityFields[0]}'"
			writeSimpleToString(equalityFields[0], equalityFieldType, writer)
		} else if (equalityFields.size() > 1) {
			writeComplexHashCode(equalityFields[0], writer)
			writeEquals(equalityFields, dtoSimpleClassName, writer)
			writeComplexToString(fields, dtoSimpleClassName, writer)
		} else {
			writeComplexToString(fields, dtoSimpleClassName, writer)
		}
				 

		// Class terminator.
		writer.write "}${eol}"

		// All done. Make sure all data has been pushed to the destination
		// before we leave.
		writer.flush()

		return relations
	}

    
    protected List getEqualityFields(Class clazz) {
        List equalityFields = []
        clazz.declaredFields.each { field ->
            if (field) {
//                println "next Field: $field.name, getAnnotation(DtoEquality.class)="+field.getAnnotation(DtoEquality.class)
//                println "            $field.name, getDeclaredAnnotations()="+field.getDeclaredAnnotations()
                if (field.getAnnotation(DtoEquality.class)) {
                    equalityFields << field.name
                }
            }
        }
        return equalityFields
    }
    
    protected List getIdOnlyFields(Class clazz) {
        List idOnlyFields = []
        clazz.declaredFields.each { field ->
            if (field) {
//                println "            $field.name, getDeclaredAnnotations()="+field.getDeclaredAnnotations()
                if (field.getAnnotation(DtoIdOnly.class)) {
                    idOnlyFields << field.name
                }
            }
        }
        println "found idOnlyFields: $idOnlyFields"
        return idOnlyFields
    }
    
    protected List getDtoImplements(Class clazz) {
        DtoImplements dtoImplements = clazz?.getAnnotation(DtoImplements.class)
        println "found DtoImplements annotation on type $clazz: types=${dtoImplements?.types()}, names=${dtoImplements?.names()}"
        def ifcTypes = (dtoImplements && dtoImplements.types()) ? (dtoImplements.types() as List) : []
        def ifcNames = (dtoImplements && dtoImplements.names()) ? (dtoImplements.names() as List) : []
        ifcNames.addAll( ifcTypes.collect { it.name } )
        return ifcNames
    }
	
    protected boolean includeVersion(Class clazz) {
        return clazz?.getAnnotation(DtoNoVersion.class) == null
    }
	protected List getIgnoredFields(Class clazz) {
		List ignoredFields = []
		clazz.declaredFields.each { field ->
			if (field) {
				if (field.getAnnotation(DtoIgnore.class)) {
					ignoredFields << field.name
				}
			}
		}
		println "found ignoredFields: $ignoredFields"
		return ignoredFields
	}

	protected List getDtoSummaryImplements(Class clazz) {
		DtoSummaryImplements dtoSummaryImplements = clazz?.getAnnotation(DtoSummaryImplements.class)
		println "found DtoSummaryImplements annotation on type $clazz: types=${dtoSummaryImplements?.types()}, names=${dtoSummaryImplements?.names()}"
		def ifcTypes = (dtoSummaryImplements && dtoSummaryImplements.types()) ? (dtoSummaryImplements.types() as List) : []
		def ifcNames = (dtoSummaryImplements && dtoSummaryImplements.names()) ? (dtoSummaryImplements.names() as List) : []
		ifcNames.addAll( ifcTypes.collect { it.name } )
		return ifcNames
	}

    protected Map getSummaryFields(Class clazz) {
        Map summaryFields = [:]
        clazz.declaredFields.each { field ->
            if (field) {
				def annotations = field.getDeclaredAnnotations()*.annotationType() 
				def summaryFieldAnnotations = annotations.findAll { it == DtoSummary.class }
                for (DtoSummary nextAnnotation in summaryFieldAnnotations) {
					String varName = nextAnnotation.name()
					if (!varName) {
						varName = field.name
					}
					String varType = (nextAnnotation.type() != DtoSummary.DEFAULT) ? nextAnnotation.type().name : null
					if (!varType) {
						varType = nextAnnotation.typeName() ? nextAnnotation.typeName() : null
					}
					if (!varType) {
						varType = field.type.name
					}

                    summaryFields << field.name
                }
            }
        }
        println "found summaryFields: $summaryFields"
        return idOnlyFields
    }
    
    protected void writeSimpleHashCode(String equalityFieldName, Writer writer) {
        writer.write eol
       writer.write "${indent}/* (non-Javadoc)${eol}"
       writer.write "${indent} * @see java.lang.Object#hashCode()${eol}"
       writer.write "${indent} */${eol}"
       writer.write "${indent}@Override${eol}"
       writer.write "${indent}public int hashCode() {${eol}"
       writer.write "${indent * 2}return ${equalityFieldName} != null ? ${equalityFieldName}.hashCode() : 0;${eol}"
       writer.write "${indent}}${eol}"
    }
    
    protected void writeComplexHashCode(List equalityFieldNames, Writer writer) {
        writer.write eol
        writer.write "${indent}/* (non-Javadoc)${eol}"
        writer.write "${indent} * @see java.lang.Object#hashCode()${eol}"
        writer.write "${indent} */${eol}"
        writer.write "${indent}@Override${eol}"
        writer.write "${indent}public int hashCode() {${eol}"
        writer.write "${indent * 2}final int prime = 31;${eol}"
        writer.write "${indent * 2}int result = 1;${eol}"
        equalityFieldNames.each { fieldName ->
            writer.write "${indent * 2}result = prime * result + ((${fieldName} == null) ? 0 : ${fieldName}.hashCode());${eol}"
        }
        writer.write "${indent * 2}return result;${eol}"
        writer.write "${indent}}${eol}"
    }
    
    
    protected void writeEquals(List equalityFieldNames, String dtoSimpleClassName, Writer writer) {
        writer.write eol
        writer.write "${indent}/* (non-Javadoc)${eol}"
        writer.write "${indent} * @see java.lang.Object#equals(java.lang.Object)${eol}"
        writer.write "${indent} */${eol}"
        writer.write "${indent}@Override${eol}"
        writer.write "${indent}public boolean equals(Object obj) {${eol}"
        writer.write "${indent*2}if (this == obj) {${eol}"
        writer.write "${indent*3}return true;${eol}"
        writer.write "${indent*2}}${eol}"
        writer.write "${indent*2}if (obj == null) {${eol}"
        writer.write "${indent*3}return false;${eol}"
        writer.write "${indent*2}}${eol}"
        writer.write "${indent*2}if (!(obj instanceof ${dtoSimpleClassName})) {${eol}"
        writer.write "${indent*3}return false;${eol}"
        writer.write "${indent*2}}${eol}"
        writer.write "${indent*2}${dtoSimpleClassName} other = (${dtoSimpleClassName}) obj;${eol}"
        equalityFieldNames.each { fieldName ->
            writer.write "${indent*2}if (${fieldName} == null) {${eol}"
            writer.write "${indent*3}if (other.${fieldName} != null) {${eol}"
            writer.write "${indent*4}return false;${eol}"
            writer.write "${indent*3}}${eol}"
            writer.write "${indent*2}} else if (!${fieldName}.equals(other.${fieldName})) {${eol}"
            writer.write "${indent*3}return false;${eol}"
            writer.write "${indent*2}}${eol}"
        }
        writer.write "${indent*2}return true;${eol}"
        writer.write "${indent}}${eol}"
    }
    
    protected void writeSimpleToString(String equalityFieldName, String equalityFieldType, Writer writer) {
        writer.write eol
        writer.write "${indent}/* (non-Javadoc)${eol}"
        writer.write "${indent} * @see java.lang.Object#toString()${eol}"
        writer.write "${indent} */${eol}"
        writer.write "${indent}@Override${eol}"
        writer.write "${indent}public String toString() {${eol}"
        if (equalityFieldType == "String") {
            writer.write "${indent * 2}return $equalityFieldName;${eol}"
        } else {
            writer.write "${indent * 2}return $equalityFieldName != null ? ${equalityFieldName}.toString() : \"\";${eol}"
        }
        writer.write "${indent}}${eol}"

    }
    
    protected void writeComplexToString(List fields, String dtoSimpleClassName, Writer writer) {
        writer.write eol
        writer.write "${indent}/* (non-Javadoc)${eol}"
        writer.write "${indent} * @see java.lang.Object#toString()${eol}"
        writer.write "${indent} */${eol}"
        writer.write "${indent}@Override${eol}"
        writer.write "${indent}public String toString() {${eol}"
        writer.write "${indent * 2}StringBuilder sb = new StringBuilder();${eol}"
        writer.write "${indent * 2}sb.append(\"${dtoSimpleClassName}[\");${eol}"
        fields.each { field ->
            if (field.idName) { 
                writer.write "${indent * 2}sb.append(\"\\n\\t${field.idName}: \" + this.${field.idName});${eol}"
            } else {
                writer.write "${indent * 2}sb.append(\"\\n\\t${field.name}: \" + this.${field.name});${eol}"
            }
        }
        writer.write "${indent * 2}sb.append(\"]\");${eol}"
        writer.write "${indent * 2}return sb.toString();${eol}"
        writer.write "${indent}}${eol}"

    }

    protected void addImportIfNecessary(List imports, String hostPackage, Class clazz, boolean isAssociation) {
        def pkg = isAssociation ? getTargetPackage(clazz.package?.name) : clazz.package?.name
        if (pkg && pkg != hostPackage && pkg != "java.lang") {
            imports << "${pkg}.${clazz.simpleName}${isAssociation ? 'DTO' : ''}"
        }
    }

    protected File getDtoFile(File rootDir, GrailsDomainClass dc, String targetPkg) {
        def pkgPath = ""
        if (targetPkg) pkgPath = targetPkg.replace(".", "/") + '/'
        return new File(rootDir, "${pkgPath}${dc.shortName}DTO.java")
    }
	
	protected File getSummaryFile(File rootDir, GrailsDomainClass dc, String targetPkg) {
		def pkgPath = ""
		if (targetPkg) pkgPath = targetPkg.replace(".", "/") + '/'
		return new File(rootDir, "${pkgPath}${dc.shortName}Summary.java")
	}

    protected String getTargetPackage(final String dcPkg) {
        def targetPkg = dcPkg ?: ""
        if (packageTransforms) {
            // Find a transform that matches the domain class package.
            // If the default package is in the transforms map (i.e.
            // the empty string is a key), the domain class package
            // must be an exact match. Otherwise, sub-packages match.
            def entry = packageTransforms.find { key, val -> key ? targetPkg?.startsWith(key) : targetPkg == key }
            if (entry) {
                // Found one, so use the associated package name as the
                // target package.
                targetPkg = targetPkg.replace(entry.key, entry.value)
            }
            else if (packageTransforms["*"]) {
                // Didn't find a matching transform, but did find the
                // wildcard one.
                targetPkg = packageTransforms["*"]
            }
        }

        return targetPkg
    }
}
